from typing import TypeVar
from types import GenericAlias


class Struct:
    ...


def parse_input(type_):
    if type(type_) != TypeVar and issubclass(type_, Struct):
        raise NotImplementedError()
        # return {"type": tuple, "components": ...}
    elif type(type_) == GenericAlias:
        args = type_.__args__
        if len(args) > 1:
            raise NotImplementedError(f"Couldn't parse {type_}")
        if type_.__origin__ != list:
            raise RuntimeWarning(f"Use list[...] instead of {type_}")

        return {"type": str(args[0]).lstrip("~") + "[]", "components": []}

    return {"type": str(type_).lstrip("~"), "components": []}



from typing import TypeVar, get_type_hints

from .events import Event
from .inputs import parse_input


class Abi:
    # https://docs.soliditylang.org/en/v0.8.13/abi-spec.html#json
    @classmethod
    def to_abi(cls):
        abi = []

        for name, type_ in cls.__annotations__.items():
            abi.append(
                {
                    "type": "function",
                    "name": name,
                    "inputs": [],
                    "outputs": [{"name": "", **parse_input(type_)}],
                    "stateMutability": "view",
                }
            )

        for item_name in set(dir(cls)) - set(dir(Abi)):
            item = getattr(cls, item_name)
            if item_name == "__init__":
                # constructor
                ...
            elif item_name == "__annotations__":
                # constants, already parsed
                ...
            elif not callable(item):
                if issubclass(item, Event):
                    # events
                    abi.append(item.to_abi_item())
            else:
                inputs = []
                outputs = []
                for name, type_ in get_type_hints(item).items():
                    if name == "return":
                        outputs = [{"name": "", **parse_input(type_)}]
                    else:
                        inputs.append({"name": name, **parse_input(type_)})

                abi.append(
                    {
                        "type": "function",
                        "name": item_name,
                        "inputs": inputs,
                        "outputs": outputs,
                        "stateMutability": getattr(
                            item, "_state_mutability", "nonpayable"
                        ),
                    }
                )

        return abi



from abinator import (
    Event,
    Abi,
    indexed,
    view,
    payable,
    address,
    uint256,
    string,
    uint8,
    uint,
    bool,
)


class ERC20(Abi):
    name: string
    symbol: string
    decimals: uint8
    totalSupply: uint256

    class Transfer(Event):
        from_: indexed(address)
        to: indexed(address)
        value: uint256

    @view
    def allowance(owner: address, spender: address) -> uint256:
        ...

    @view
    def balanceOf(account: address) -> uint256:
        ...

    def approve(spender: address, amount: uint256) -> bool:
        ...

    def transfer(to: address, amount: uint256) -> bool:
        ...

    def transferFrom(from_: address, to: address, amount: uint256) -> bool:
        ...


class WETH(ERC20):
    @payable
    def deposit():
        ...

    def withdraw(wad: uint):
        ...


print(ERC20().to_abi())
